/* drivers/power/bcm59040_battery.c
 *
 * Copyright (c) 2009 TomTom BV <http://www.tomtom.com>
 * Copyright (c) 2009 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 */

#include <linux/workqueue.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/power_supply.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <asm/atomic.h>
#include <asm/arch/bcm_gen_battery.h>
#include <linux/delay.h>
#include <linux/notifier.h>
#include <plat/fdt.h>
#include <linux/broadcom/bcm_adc.h>
#include <linux/broadcom/pmu_chip.h>

#define PFX "battery: "

#define BAT_NUM_SAMPLES_PER_AVERAGE	8

#define BAT_MIN_VOLTAGE	3000000
#define BAT_MAX_VOLTAGE	4500000


/* Debug logging */
#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG 0

#define DBG_ERROR	0x01
#define DBG_INFO	0x02
#define DBG_TRACE	0x04
#define DBG_TRACE2	0x08
#define DBG_DATA	0x10
#define DBG_DATA2	0x20


#if DEBUG

#define DBG_DEFAULT_LEVEL	DBG_INFO | DBG_TRACE

static int 		gLevel = DBG_DEFAULT_LEVEL;

#define BATT_DEBUG(level,x) {if (level & gLevel) printk x;}
#else
#define BATT_DEBUG(level,x)
#endif

extern atomic_t bcm59040_batt_status;
struct atomic_adc_polled_average_value
{
	spinlock_t	lock;
	u64		running_total;
	s32		the_samples[BAT_NUM_SAMPLES_PER_AVERAGE];
	u32		num_sample;
	s32		current_average;
	u8    		buffer_wrapped;
};

struct bcm_power_supply
{
	struct power_supply			psy;

	struct bcm_battery_platform_info			*platform_info;

	struct atomic_adc_polled_average_value		avg_batt_volt;
	struct atomic_adc_polled_average_value		avg_batt_curr;
	struct atomic_adc_polled_average_value		avg_charge_curr;

	atomic_t					is_charging;
	atomic_t					current_capacity;
	atomic_t					remaining_time;
	atomic_t					current_consumption;
	atomic_t					running;

	struct delayed_work			work;
	struct workqueue_struct*	workqueue;
};

static int battery_get_property(
		struct power_supply *psy,
		enum power_supply_property psp,
		union power_supply_propval *val);
static u32 battery_get_charge(
		struct bcm_power_supply *psy,
		u32 result_as_percent);

static enum power_supply_property	bcm_battery_properties[] =
{
	POWER_SUPPLY_PROP_HEALTH,
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_PRESENT,
	POWER_SUPPLY_PROP_ONLINE,
	POWER_SUPPLY_PROP_VOLTAGE_NOW,
	POWER_SUPPLY_PROP_VOLTAGE_AVG,
	POWER_SUPPLY_PROP_CURRENT_NOW,
	POWER_SUPPLY_PROP_CURRENT_AVG,
	POWER_SUPPLY_PROP_CHARGE_FULL_DESIGN,
	POWER_SUPPLY_PROP_CHARGE_EMPTY_DESIGN,
	POWER_SUPPLY_PROP_CHARGE_NOW,
	POWER_SUPPLY_PROP_CHARGE_AVG,
	POWER_SUPPLY_PROP_CAPACITY,
	POWER_SUPPLY_PROP_TIME_TO_EMPTY_AVG
};

static struct bcm_power_supply	bcm59040_battery_psy=
{
	.psy =
	{
		.name				= "battery",
		.type				= POWER_SUPPLY_TYPE_BATTERY,
		.properties			= bcm_battery_properties,
		.num_properties		= ARRAY_SIZE(bcm_battery_properties),
		.get_property		= battery_get_property,
		.use_for_apm		= 1,
	},
	.platform_info			= NULL,
	.avg_batt_volt =
	{
		.lock				= SPIN_LOCK_UNLOCKED,
		.running_total		= 0,
		.the_samples        = {},
		.num_sample			= 0,
		.buffer_wrapped     = 0,
	},
	.avg_batt_curr =
	{
		.lock				= SPIN_LOCK_UNLOCKED,
		.running_total		= 0,
		.the_samples		= {},
		.num_sample 		= 0,
		.buffer_wrapped 		= 0,
	},
	.avg_charge_curr =
	{
		.lock				= SPIN_LOCK_UNLOCKED,
		.running_total		= 0,
		.the_samples		= {},
		.num_sample 		= 0,
		.buffer_wrapped 	= 0,
	},
	.is_charging			= ATOMIC_INIT(0),
	.current_capacity		= ATOMIC_INIT(0),
	.remaining_time			= ATOMIC_INIT(0),
	.current_consumption   	= ATOMIC_INIT(350),
	.running				= ATOMIC_INIT(0),
};

static ssize_t battery_show_current(struct device *dev,
									struct device_attribute *attr, char *buffer)
{
	snprintf(buffer, PAGE_SIZE, "%d",
			 atomic_read(&bcm59040_battery_psy.current_consumption));
	return strlen(buffer);
}

static ssize_t battery_store_current(struct device *dev,
									 struct device_attribute *attr,
									 const char *buffer, size_t size)
{
	long new_current_consumption = simple_strtol(buffer, NULL, 10);
	if (new_current_consumption)
	{
		atomic_set(&bcm59040_battery_psy.current_consumption,
				   new_current_consumption);
		return size;
	}
	return -EINVAL;
}

static DEVICE_ATTR(current_consumption, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP,
				   battery_show_current, battery_store_current);


static inline void update_average_atomic_adc_value(
		struct atomic_adc_polled_average_value *arg)
{
	u32 i;
	s32 result = 0;
	u32 sample_count;
	
	sample_count = (arg->buffer_wrapped) ? 
		BAT_NUM_SAMPLES_PER_AVERAGE : arg->num_sample;
	for (i = 0; i < sample_count; i++)
	{
		result += arg->the_samples[i];
	}
	if(sample_count != 0)
		result /= (s32)sample_count;
	else
		result = 0;
	arg->current_average = result;
}

static inline void reset_atomic_average(struct atomic_adc_polled_average_value *arg)
{
	unsigned long flags;

	spin_lock_irqsave(&(arg->lock), flags);
	arg->num_sample = 0;
	arg->buffer_wrapped = 0;
	arg->running_total  = 0;
	arg->current_average = 0;
	memset(arg->the_samples, 0, BAT_NUM_SAMPLES_PER_AVERAGE);
	spin_unlock_irqrestore(&(arg->lock), flags);
}

static inline void add_atomic_adc_value(struct atomic_adc_polled_average_value *arg, s32 value)
{
	unsigned long flags;

	spin_lock_irqsave(&(arg->lock), flags);
	arg->the_samples[(arg->num_sample++) % BAT_NUM_SAMPLES_PER_AVERAGE] = value;
	if (arg->num_sample == BAT_NUM_SAMPLES_PER_AVERAGE)
	{
		arg->buffer_wrapped = 1;
	}
	update_average_atomic_adc_value(arg);
	spin_unlock_irqrestore(&(arg->lock), flags);
}

static inline s32 get_average_atomic_adc_value(struct atomic_adc_polled_average_value *arg)
{
	unsigned long flags;
	s32 retval;

	spin_lock_irqsave(&(arg->lock), flags);
	retval = arg->current_average;
	spin_unlock_irqrestore(&(arg->lock), flags);

	return retval;
}

static inline u32 get_last_atomic_adc_value(struct atomic_adc_polled_average_value *arg)
{
	unsigned long flags;
	u32 retval;

	spin_lock_irqsave(&(arg->lock), flags);
	retval = arg->the_samples[(arg->num_sample-1) % BAT_NUM_SAMPLES_PER_AVERAGE];
	spin_unlock_irqrestore(&(arg->lock), flags);

	return retval;
}

/* Return: y0 */
static inline long linear_interpolate(long x0,  long x1, long y1, long x2, long y2)
{
	int64_t y0;
	long div, sign = 0;

	if(x1 == x2)
		return y1;
	y0 = (x0 - x1);
	y0 *= (y2 - y1);
	if (y0 < 0)
	{
		y0 = -y0;
		sign = !sign;
	}
	div = x2 - x1;
	if (div < 0)
	{
		div = -div;
		sign = !sign;
	}

	do_div(y0, div); // This works only for positive numbers, tweak needed

	if(sign)
		y0 = -y0;

	y0 += y1;

	return (long)y0;
}

/*
 * Get remaining capacity from a given discharge curve, using linear
 * interpolation.
 * Parameters: curve - discharge curve to use
 *             result_as_percent - 1 : return result as % of capacity
 *                                 0 : return result as remaining time in sec
 */
#define GET_CAPACITY(curve) \
	 (result_as_percent ? curve.remaining_capacity : curve.remaining_time)
static u32 get_capacity_from_curve(batt_discharge_curve_t *curve, u32 voltage,
		int result_as_percent)
{
	batt_discharge_lut_t	*discharge_lut;
	u32			i, cap1, cap2, voltage1, voltage2;

	cap1 = cap2 = voltage1 = voltage2 = (u32)-1;
	discharge_lut = curve->lut;
	for (i = 0; i < curve->dclut; i++)
	{
		if (voltage > discharge_lut[i].vbatt)
		{
			voltage1 = discharge_lut[i].vbatt;
			cap1 = GET_CAPACITY(discharge_lut[i]);
			if (i > 0)
			{
				voltage2 = discharge_lut[i - 1].vbatt;
				cap2 = GET_CAPACITY(discharge_lut[i - 1]);
			}
			break;
		}
	}
	if (cap1 == (u32)-1)
	{
		voltage1 = discharge_lut[curve->dclut - 1].vbatt;
		cap1 = GET_CAPACITY(discharge_lut[curve->dclut - 1]);
	}
	if (cap2 == (u32)-1)
	{
		cap2 = cap1;
		voltage2 = voltage1;
	}
	return linear_interpolate(voltage, voltage1, cap1, voltage2, cap2);
}
#undef GET_CAPACITY

/* 
 * Converts the battery voltage into a capacity measurement using
 * precalculated look up tables: we first select the correct 2 tables
 * based on current consumption, then evaluate capacity/remaining time
 * based on battery voltage.
 * 
 * battery_volts:     Current battery voltage
 * result_as_percent: 1 - the capacity is returned as a percentage,
 *                    0 -  "   in minutes.
 */
static u32 convert_volts_to_capacity(u32 battery_volts, int result_as_percent)
{
	batt_discharge_lut_t	*curve1, *curve2, *curves;
	struct bcm_batt_info	*batt_info;
	u32						i, cap1, cap2;

	batt_info = bcm59040_battery_psy.platform_info->battery_info;

	if(batt_info->battery.discharge_curve_entries == 0)
		return 0;

	curves = batt_info->battery.discharge_curves;

	/* Find the two nearest curves */
	curve1 = curve2 = NULL;
	for (i = 0; i < batt_info->battery.discharge_curve_entries; i++)
	{
		if (battery_volts > curves[i].vbatt)
		{
			curve1 = &(curves[i]);
			if(i > 0)
				curve2 = &(curves[i - 1]);
			break;
		}
	}

	if(curve1 == NULL)
		curve1 = &(curves[batt_info->battery.discharge_curve_entries - 1]);

	if(result_as_percent)
		cap1 = curve1->remaining_capacity;
	else
		cap1 = curve1->remaining_time;

	if (curve2)
	{
		if(result_as_percent)
			cap2 = curve2->remaining_capacity;
		else
			cap2 = curve2->remaining_time;
	}
	else
	{
		cap2 = cap1;
		curve2 = curve1;
	}

	/* Do a linear interpolation between the two values */
	return linear_interpolate(battery_volts,
							  curve1->vbatt, cap1,
							  curve2->vbatt, cap2);
}

/* R1 = TOP resistor, R2 = BOTTOM resistor. */
/* Return in uV */
static u32 batteryadc2voltage(struct bcm_battery_platform_info *bcm_plat_info)
{
	u32 sample;

	sample = (u32)(bcm_plat_info->get_battery_voltage_adc());
	sample *= 1000000;		/* keep in uV */
	sample >>=10; // divide by 1024
	sample *= 6;

	return sample;
}

/* Returns in uA */
static s32 batteryadc2current(struct bcm_battery_platform_info *bcm_plat_info, u32 *charging)
{
	s32 sample;
	u64 hpval;

	*charging=0;

	sample = -(s32)bcm_plat_info->get_fuel_gauge_sample();
	if( sample < 0 ){/*current is going into the battery so charging*/
		sample=-sample;/*do_div only works with u64*/
		*charging=1;
	}
	hpval=((u64) sample) * 976500;
	hpval*=1000000;
//      do_div(hpval, bcm_plat_info->battery_info->adc.charge_current_resistor);
        do_div(hpval, 1000000000);
	sample=(s32) hpval;
	if(*charging) sample=-sample;

	return sample;
}

static int battery_has_samples(struct bcm_power_supply *bcm_psy)
{
	unsigned long flags;
	int has_samples;
	struct atomic_adc_polled_average_value *arg;

	arg = &bcm_psy->avg_batt_volt;
	spin_lock_irqsave(&(arg->lock), flags);
	has_samples = !!arg->num_sample;
	spin_unlock_irqrestore( &(arg->lock), flags );

	return has_samples;
}

#define CHARGING_NOT_INITIALIZED 0xA5

static void battery_poll(struct work_struct *work)
{
	static u32 is_charging_prev = CHARGING_NOT_INITIALIZED;

	struct bcm_power_supply	*bcmpsy = &bcm59040_battery_psy;
	u32 battery_voltage, averaged_batt_voltage;
	s32 battery_current, averaged_batt_current;
	u32 previous_cap, cap, rem;
	u32 is_charging;
	int charger_id;
	u8 changed = 0;


    BATT_DEBUG(DBG_TRACE,( "Entering battery_poll\n" ));

	/* Get the current. */
	battery_current = batteryadc2current(bcmpsy->platform_info, &is_charging);
	add_atomic_adc_value(&bcmpsy->avg_batt_curr, battery_current);
	averaged_batt_current = get_average_atomic_adc_value(&bcmpsy->avg_batt_curr);

	/* Get the voltage. */
	battery_voltage = batteryadc2voltage(bcmpsy->platform_info);

	if (    (battery_voltage > BAT_MIN_VOLTAGE)
	     && (battery_voltage < BAT_MAX_VOLTAGE)) {
		add_atomic_adc_value(&bcmpsy->avg_batt_volt, battery_voltage);
	}

	averaged_batt_voltage = (u32)get_average_atomic_adc_value(&bcmpsy->avg_batt_volt);

    BATT_DEBUG(DBG_TRACE,( "vm_bat = %u\n", battery_voltage ));

	/* We compute capacity at every ~1 sec poll and cache the value
	 * for client to read for performance reasons. */
	previous_cap = atomic_read(&bcmpsy->current_capacity);
	cap = convert_volts_to_capacity(averaged_batt_voltage, 1);

    BATT_DEBUG(DBG_TRACE,( "cap = %d\n", cap ));

	/* We only test for % charge change, remaining time will change
	 * almost always and we don't want to update every time. Also
	 * we re-compute remaining time only if capacity changed. */
	if(cap != previous_cap)
	{
		changed = 1;
		atomic_set(&bcmpsy->current_capacity, cap);
		rem = convert_volts_to_capacity(averaged_batt_voltage, 0);

	    BATT_DEBUG(DBG_TRACE,( "rem = %d\n", rem ));

		atomic_set(&bcmpsy->remaining_time, rem);
	}

	/* See if we're currently charging */
	is_charging_prev = atomic_read(&(bcmpsy->is_charging));
	
	if(is_charging != is_charging_prev)
	{
		changed = 1;
		atomic_set(&(bcmpsy->is_charging), is_charging);
		is_charging_prev = is_charging; 
		if (is_charging)
		{
			reset_atomic_average(&bcm59040_battery_psy.avg_charge_curr);
		}
		else
		{
			reset_atomic_average(&bcm59040_battery_psy.avg_batt_volt);
			reset_atomic_average(&bcm59040_battery_psy.avg_batt_curr);
		}
	}

	if(is_charging)
		add_atomic_adc_value(&bcmpsy->avg_batt_curr, battery_current);
	else
		add_atomic_adc_value(&bcmpsy->avg_charge_curr, battery_current);

	if (changed)
		power_supply_changed(&bcmpsy->psy);

	if(atomic_read(&bcmpsy->running))
	{
		/* Poll faster until we have at least
		 * BAT_NUM_SAMPLES_PER_AVERAGE samples */
		queue_delayed_work(bcmpsy->workqueue, &bcmpsy->work, 
				BATT_POLL_INTERVAL);
	}
}

static int battery_probe(struct platform_device *pdev)
{
	struct bcm_battery_platform_info	*bcm_platinfo =
		(struct bcm_battery_platform_info *)pdev->dev.platform_data;

	/* Initialize platform dependant stuff. */
	if(bcm_platinfo->platform_init(pdev))
		return -EINVAL;

	/* Register the battery. */
	bcm59040_battery_psy.platform_info = bcm_platinfo;
	if(power_supply_register(&pdev->dev, &bcm59040_battery_psy.psy))
		return -EINVAL;

	/* Setup the workqueue. Needs to be done before calling
	   battery_state_changed */
	bcm59040_battery_psy.workqueue = create_workqueue(BCM_BATTERY_DRIVER);
	INIT_DELAYED_WORK(&bcm59040_battery_psy.work, battery_poll);

	/* Sanity check: if Vmaxf > Vmax and/or Vminf > Vmin then the unit would
	 * suspent too urly because the interpolated
	 * Voltage would be lower then the actual Vbatt
	 */
	if(bcm_platinfo->battery_info->battery.Vmaxf > bcm_platinfo->battery_info->battery.Vmax)
		bcm_platinfo->battery_info->battery.Vmaxf = bcm_platinfo->battery_info->battery.Vmax;

	if(bcm_platinfo->battery_info->battery.Vminf > bcm_platinfo->battery_info->battery.Vmin)
		bcm_platinfo->battery_info->battery.Vminf = bcm_platinfo->battery_info->battery.Vmin;


	if (device_create_file(&pdev->dev, &dev_attr_current_consumption))
	{
		printk(KERN_ERR PFX
			"unable to create the 'current_consumption' "
			"attribute of the battery driver.\n");
		power_supply_unregister(&bcm59040_battery_psy.psy);
		return -EINVAL;
	}
		
	/* Reset the VBatt averaging counters */
	reset_atomic_average(&bcm59040_battery_psy.avg_batt_volt);
	reset_atomic_average(&bcm59040_battery_psy.avg_batt_curr);
	reset_atomic_average(&bcm59040_battery_psy.avg_charge_curr);
	
	/* Start up the battery polling work Q asap. */
	atomic_set(&bcm59040_battery_psy.running, 1);
	queue_work(bcm59040_battery_psy.workqueue, &bcm59040_battery_psy.work);
	
	return 0;
}

static int battery_remove(struct platform_device *pdev)
{
	struct bcm_battery_platform_info	*bcm_platinfo =
		(struct bcm_battery_platform_info *)pdev->dev.platform_data;

	/* Cancel polling work Q. */
	atomic_set(&bcm59040_battery_psy.running, 0);
	cancel_delayed_work_sync(&bcm59040_battery_psy.work);
	destroy_workqueue(bcm59040_battery_psy.workqueue);

	power_supply_unregister(&bcm59040_battery_psy.psy);

	bcm_platinfo->platform_remove(pdev);
	return 0;
}

static u32 battery_get_charge(struct bcm_power_supply *bcmpsy,
		u32 result_as_percent)
{
	if (!battery_has_samples(bcmpsy))
	{
		printk(KERN_DEBUG PFX "%s: waiting for samples\n",
				__FUNCTION__);
		do
		{ 
			msleep(10); 
		} while (!battery_has_samples(bcmpsy));
		printk(KERN_DEBUG PFX "%s: samples found\n", __FUNCTION__);
	}

	if(result_as_percent)
		return atomic_read(&bcmpsy->current_capacity);
	else
		return atomic_read(&bcmpsy->remaining_time);
}


static int battery_get_property(struct power_supply *psy,
		enum power_supply_property psp,
		union power_supply_propval *val )
{
	u32 chg_current;
	struct bcm_power_supply *bcmpsy = (struct bcm_power_supply *)container_of(
		psy, struct bcm_power_supply, psy);

	switch(psp)
	{
		case POWER_SUPPLY_PROP_TIME_TO_EMPTY_AVG:
			if (atomic_read(&bcmpsy->is_charging))
			{
				/* time is invalid while charging */
				val->intval = 1;
			}
			else
			{
				val->intval = battery_get_charge(bcmpsy, 0);
			}
			break;

		/* Convert this percentage for CHARGE_NOW and CHARGE_AVG to capacity in uAh */			
		case POWER_SUPPLY_PROP_CHARGE_NOW:
			val->intval = convert_volts_to_capacity(get_last_atomic_adc_value(&bcmpsy->avg_batt_volt), 1);
			break;

		case POWER_SUPPLY_PROP_CHARGE_AVG:
			val->intval = convert_volts_to_capacity(get_average_atomic_adc_value(&bcmpsy->avg_batt_volt), 1);
			break;

		case POWER_SUPPLY_PROP_CAPACITY:/* added to support APM code */
			val->intval = battery_get_charge(bcmpsy, 1);
			break;

		/* Convert CHARGE_FULL/EMPTY to capacity in uAh */
		case POWER_SUPPLY_PROP_CHARGE_FULL_DESIGN:
			val->intval = 100;
			break;
		case POWER_SUPPLY_PROP_CHARGE_EMPTY_DESIGN:
			val->intval = 0;
			break;

		case POWER_SUPPLY_PROP_PRESENT :
			val->intval = 1;
			break;
		case POWER_SUPPLY_PROP_ONLINE :
			val->intval = 1;
			break;

		case POWER_SUPPLY_PROP_HEALTH :
			val->intval = atomic_read( &bcm59040_batt_status );
			break;

		case POWER_SUPPLY_PROP_STATUS :
			if (atomic_read(&bcmpsy->is_charging))
			{
				/* Real test is if(chg current < n) where n
				 * is some small value, this will prevent
				 * charging dragging on for hours */
				chg_current = get_average_atomic_adc_value(
					&bcmpsy->avg_charge_curr);
				if (chg_current < 10)	
					val->intval = POWER_SUPPLY_STATUS_FULL;
				else
				if(chg_current < 10)	
					val->intval = POWER_SUPPLY_STATUS_FULL;
				else
				if( chg_current < 0 )
					val->intval = POWER_SUPPLY_STATUS_DISCHARGING;
				else if( chg_current < 10 )
					val->intval = POWER_SUPPLY_STATUS_FULL;
				else
					val->intval = POWER_SUPPLY_STATUS_CHARGING;
			}
			else
			{
				val->intval = POWER_SUPPLY_STATUS_DISCHARGING;
			}
			break;

		case POWER_SUPPLY_PROP_VOLTAGE_NOW:
			val->intval = get_last_atomic_adc_value(&bcmpsy->avg_batt_volt);
			break;

		case POWER_SUPPLY_PROP_VOLTAGE_AVG:
			val->intval = get_average_atomic_adc_value(&bcmpsy->avg_batt_volt);
			break;

		case POWER_SUPPLY_PROP_CURRENT_NOW:
			val->intval = get_last_atomic_adc_value(&bcmpsy->avg_batt_curr);
			break;

		case POWER_SUPPLY_PROP_CURRENT_AVG:
			val->intval = get_average_atomic_adc_value(&bcmpsy->avg_batt_curr);
			break;

		default :
			return -EINVAL;
	}

	return 0;
}

#ifdef CONFIG_PM
static int bcm59040_battery_suspend(struct device *dev)
{
	struct platform_device *pdev;

	pdev = to_platform_device(dev);

	atomic_set(&bcm59040_battery_psy.running, 0);
	cancel_delayed_work_sync(&bcm59040_battery_psy.work);
	flush_workqueue(bcm59040_battery_psy.workqueue);

	BATT_DEBUG(DBG_TRACE,( "Suspend processed for %s.\n", pdev->name ));

	return 0;
} /* bcm59040_battery_suspend */

static int bcm59040_battery_resume(struct device *dev)
{
	struct platform_device *pdev;

	pdev = to_platform_device(dev);

	queue_delayed_work(bcm59040_battery_psy.workqueue,
				&bcm59040_battery_psy.work, BATT_POLL_INTERVAL);
	atomic_set(&bcm59040_battery_psy.running, 1);

	BATT_DEBUG(DBG_TRACE,( "Resume processed for %s.\n", pdev->name ));

	return 0;
} /* bcm59040_battery_resume */

#else
#define bcm59040_battery_suspend NULL
#define bcm59040_battery_resume NULL
#endif

static struct pm_ext_ops bcm59040_battery_pm_ops =
{
    .base =
    {
        .suspend    = bcm59040_battery_suspend,
        .resume     = bcm59040_battery_resume
    }
};

static struct platform_driver bcm59040_battery_driver =
{
	.driver =
	{
		.name	= BCM_BATTERY_DRIVER,
	},
	.pm		= &bcm59040_battery_pm_ops,
	.probe	= battery_probe,
	.remove	= battery_remove,
};

static int __init bcm59040_battery_init(void)
{
	return platform_driver_register(&bcm59040_battery_driver);
}

static void __exit bcm59040_battery_exit(void)
{
	platform_driver_unregister(&bcm59040_battery_driver);
}

module_init(bcm59040_battery_init);
module_exit(bcm59040_battery_exit);

MODULE_DESCRIPTION("BCM59040 Battery Driver");
MODULE_VERSION("2.0");
MODULE_LICENSE("GPL");
